/* Original code from Mike Godwin at
http://www.mat.ucsb.edu/~g.legrady/academic/courses/06w259/projs/mg/KohonenSpectrum/
Adapted by Kitty Currier, 2/2014 */

class KohonenSet{
  float learningRate= .25;
  int numNodes, numWeights;
  int vNodes, hNodes;
  float MAX_ITERATIONS = 250;
  int iterationCounter = 0;
  float neighborDist;
  float weightArray[][]; //an array of 'node' by 'weight' dimensions
  float kset[][];  //values to be sorted
  Integer ksetRandomOrder[];
  int mapMode = 0;

  void KohonenSet() {
  }
  
  void initWeights(int hNodes, int vNodes, int numWeights) {
    weightArray = new float[vNodes*hNodes][numWeights];
    numNodes = vNodes*hNodes;
    this.hNodes = hNodes;
    this.vNodes = vNodes;
    this.numWeights = numWeights;
    for (int i=0; i<numNodes; i++){
      for (int j=0; j<numWeights; j++){
        weightArray[i][j] = random(1);  // initialize random weights between 0 (inclusive) and 1 (exclusive)
//        println("weight = "+weightArray[i][j]);
      }
    }
    ksetRandomOrder = new Integer[kset.length];
    for (int i=0; i<kset.length; i++){
      ksetRandomOrder[i] = new Integer(i);
    }
  }

  void train(){
    Collections.shuffle(Arrays.asList(ksetRandomOrder));    
    if (iterationCounter<MAX_ITERATIONS){
//      neighborDist = 200*exp(-1*iterationCounter/MAX_ITERATIONS);
    neighborDist = 30+100-(100*(iterationCounter/MAX_ITERATIONS));
//      neighborDist = 32+(height/4*(1-(iterationCounter/MAX_ITERATIONS)));  
      for (int k=0; k<kset.length; k++){
        int bmu = bestMatchingUnit(ksetRandomOrder[k].intValue());
        for (int node=0; node<numNodes; node++){
          float nodeX = screenW/hNodes*(node%hNodes);
          float nodeY = screenH/vNodes*(node/hNodes);
          float bmuX = screenW/hNodes*(bmu%hNodes);
          float bmuY = screenH/vNodes*(bmu/hNodes);
          float nodeDist = constrain(dist(nodeX,nodeY,bmuX, bmuY),0,neighborDist);
          for (int i=0; i<numWeights; i++){
            if (nodeDist == 0){
              weightArray[node][i] = kset[ksetRandomOrder[k].intValue()][i];
            }
            else {
              float percent = learningRate*(1-(nodeDist/neighborDist));
              weightArray[node][i] = (1-percent)*weightArray[node][i] + percent*kset[ksetRandomOrder[k].intValue()][i];
            }
          }
        }
      }
    } 
    iterationCounter++;
  }

  int bestMatchingUnit(int inputNode){
    int bmu=0; //arbitrary
    float bestFitDistance = 1000; //arbitrarily high
    for (int node=0; node<numNodes; node++){
      float sum = 0;
      for (int i=0; i<numWeights; i++){
        sum += sq(kset[inputNode][i]-weightArray[node][i]);
      }
      float distance = sqrt(sum);
      if (distance <= bestFitDistance){
        bestFitDistance = distance;
        bmu = node;
      }
    }
    return(bmu);
  }

  void map(){
    for (int node=0; node<numNodes; node++){
      pushMatrix();
      translate(screenW/hNodes*(node%hNodes),screenH/vNodes*(node/hNodes));
      /* 
      HSB ranges:
      H: 0-240
      S: 10-90
      B: 25-75
      */  
      colorMode(HSB, 360, 100, 100);
      fill(int(weightArray[node][0]*240),int(weightArray[node][1]*80+10),int(weightArray[node][2]*50+25));
      rect(0,0,screenW/hNodes,screenW/vNodes);
      popMatrix();
    }
  }

  void errorMap(){
    for (int node=0; node<numNodes; node++){
      float nodeValue=0;
      for (int i=0; i<numWeights; i++){
        float adjacentWeights=0;
        int numWeightsChecked=0;
        if(node%hNodes != 0){
          adjacentWeights += weightArray[node-1][i]; //if you're along the left side of the screen don't check node to left
          numWeightsChecked++;
        }
        if(node%hNodes != hNodes-1){
          adjacentWeights += weightArray[node+1][i];//if you're along the right don't check to the right
          numWeightsChecked++;
        }
        if(node>=hNodes){
          adjacentWeights += weightArray[node-hNodes][i];//if you're along the top don't check above you
          numWeightsChecked++;
        }
        if(node<(numNodes-hNodes)){
          adjacentWeights += weightArray[node+hNodes][i];//if you're along the bottem don't check below yourself
          numWeightsChecked++;
        }
        //        nodeValue += weightArray[node][i];
        nodeValue += abs(weightArray[node][i] - (adjacentWeights/numWeightsChecked));
      }
      //      fill(255-abs(nodeValue/numWeights - adjacentWeights/numWeightsChecked));
      colorMode(RGB, 255);
      fill(255-5*nodeValue);
      pushMatrix();
      translate(screenW/hNodes*(node%hNodes),screenH/vNodes*(node/hNodes));
      rect(0,0,screenW/hNodes,screenH/vNodes);
      popMatrix();
    }
    colorMode(HSB, 360, 100, 100);
  }

  void listMap(){
    float m=.5; //a multiplier for easier visualization
    for(int node=0; node<numNodes; node++){
      pushMatrix();
      translate(0,screenH/vNodes*(node/hNodes));
      fill(0,99,99);
      rect(0,0,m*weightArray[node][0],screenH/vNodes);
      fill(119, 99, 99);
      rect(m*weightArray[node][0],0,m*(weightArray[node][1]),screenH/vNodes);
      fill(239,99,99);
      rect(m*(weightArray[node][0]+weightArray[node][1]),0,m*(weightArray[node][2]),screenH/vNodes);
      popMatrix();
    }
  }
  
  void listMapTwo(){
    background(230);
    float m=.5; //a multiplier for easier visualization
    for(int node=0; node<numNodes; node++){
      pushMatrix();
      translate(0,screenH/vNodes*(node/hNodes));
      fill(0,99,99);
      rect(0,0,m*weightArray[node][0],screenH/vNodes);
      fill(119, 99, 99);
      rect(150,0,m*(weightArray[node][1]),screenH/vNodes);
      fill(239,99,99);
      rect(300,0,m*(weightArray[node][2]),screenH/vNodes);
      popMatrix();
    }
  }

    void legend(){
      colorMode(HSB, 360, 100, 100);
      for (int i = 1; i<dataList.length; i++) { 
        stroke(0, 0, 85);
      /* 
        HSB ranges:
        H: 0-240
        S: 10-90
        B: 25-75
      */  
        int h = int(dataList[i][0]*240);
        int s = int(dataList[i][1]*80+10);
        int b = int(dataList[i][2]*50+25);
        fill(h, s, b);
        rect((i-1)*screenW/(dataTable.getRowCount()-1), screenH+15, screenW/(dataTable.getRowCount()-1), 10);
      }  
     
      // Text
      fill(0, 0, 99); // HSB mode, white
      textFont(font12);
      text("HUE", 20, screenH+45);
      text("No. of bar codes", 20, screenH+65);
      text("4", 35, screenH+80);
      text("3", 35, screenH+95);
      text("2", 35, screenH+110);
      
      text("SATURATION", 140, screenH+45);
      text("No. of check-outs", 140, screenH+65);
      text("Many", 155, screenH+80);
      text("Few", 155, screenH+110);
      
      text("BRIGHTNESS", 260, screenH+45);
      text("Book", 275, screenH+80);
      text("Not book", 275, screenH+110);
   
      textFont(font24);
      text("Dimensions: "+inputDims, 390, screenH+80);
      text("Records: "+dataRowsLoaded, 390, screenH+110);   
            
      // Hue color boxes
      fill(240, 40, 60);
      rect(20, screenH+70, 10, 10);
      fill(120, 40, 60);
      rect(20, screenH+85, 10, 10);
      fill(0, 40, 60);
      rect(20, screenH+100, 10, 10);
      
      // Saturation color box
      noStroke();
      colorMode(HSB, 360, 100,100);
      pushMatrix();
      translate(140, screenH+70);
      for(int i=0; i<10; i++) {
        for(int j = 0; j<40; j++) {
          stroke(120, 90-2*j, 40);  //gradient  saturation values: 10-90
          point(i,j); 
        }
      }
      popMatrix(); 
      stroke(0, 0, 15);
      noFill();
      rect(140, screenH+70, 10, 40);
      
      // Brightness color boxes
      fill(120, 40, 60); // xxx brightness range: 25, 75
      rect(260, screenH+70, 10, 10);
      fill(120, 40, 25);
      rect(260, screenH+100, 10, 10);
  }

  void bmuLabels(){
    //dots for the BMUs
    for (int i=0; i<kset.length; i++){
      int bmu = bestMatchingUnit(i);
      pushMatrix();
      translate((screenW/(2*hNodes))+screenW/hNodes*(bmu%hNodes),(screenH/(2*vNodes))+screenH/vNodes*(bmu/hNodes));
      noStroke();
      fill(0,0,99);
      ellipse(0, 0, 3, 3);
      popMatrix();
    }   
    stroke(0, 0, 15);
  }

  String toString(){
    for (int node=0; node<kset.length; node++){
      print("node: " + node + "\t value: ");
      for (int i=0; i<numWeights; i++){
        print(kset[node][i] + "\t");
      }
      println();
    }
    return("");
  }
}
